import 'dart:typed_data';

import 'package:meta/meta.dart';
import 'package:pdf/pdf.dart';
import 'package:pdf/widgets.dart';
import 'package:image/image.dart' as img;

import 'generator_card.dart';

/// Creates a PDF document from GeneratorCard lists.
class Generator {
  static const double mm = PdfPageFormat.mm;
  static const double pageHeight = 297 * mm;
  static const double pageWidth = 210 * mm;
  static const double cardHeight = 3.5 * PdfPageFormat.inch;
  static const double cardWidth = 2.5 * PdfPageFormat.inch;
  static const PdfPageFormat a4Format = PdfPageFormat(pageWidth, pageHeight);
  static const PdfPageFormat singleCard = PdfPageFormat(cardWidth, cardHeight);

  Font fontRegular;
  Font fontSemiBold;
  Font fontBold;
  TextStyle styleTitle;
  TextStyle styleType;
  TextStyle styleId;
  TextStyle styleText;
  TextStyle styleKeyword;
  TextStyle styleStats;

  String title;
  Map<GeneratorElement, img.Image> elementImages;
  Map<GeneratorElement, img.Image> frameImages;
  List<GeneratorCard> cards = <GeneratorCard>[];

  /// Assets object links images used in the cards.
  Generator(
      {@required this.title,
      @required ByteData fontDataRegular,
      @required Map<String, img.Image> elements,
      @required Map<String, img.Image> frames,
      ByteData fontDataSemiBold,
      ByteData fontDataBold}) {
    fontRegular = Font.ttf(fontDataRegular);
    fontBold = fontDataBold == null ? fontRegular : Font.ttf(fontDataBold);
    fontSemiBold =
        fontDataSemiBold == null ? fontBold : Font.ttf(fontDataSemiBold);

    elementImages = elements
        .map((var key, var entry) => MapEntry(_elementFromString(key), entry));
    frameImages = frames
        .map((var key, var entry) => MapEntry(_elementFromString(key), entry));

    styleTitle = TextStyle(font: fontRegular, fontSize: 12);
    styleType = TextStyle(font: fontRegular, fontSize: 8);
    styleId = TextStyle(font: fontRegular, fontSize: 7);
    styleText = TextStyle(font: fontRegular, lineSpacing: 1, fontSize: 7);
    styleKeyword = TextStyle(font: fontSemiBold, lineSpacing: 1, fontSize: 7);
    styleStats = TextStyle(font: fontBold, fontSize: 14);
  }

  /// Add a single card to the generator list.
  void addCard({
    @required String name,
    @required String id,
    @required String type,
    @required String stats,
    @required String text,
    @required String element,
    @required Map<String, int> cost,
    @required img.Image image,
  }) {
    var costMap = cost
        .map((var key, var entry) => MapEntry(_elementFromString(key), entry));

    cards.add(GeneratorCard(
      name: name,
      id: id,
      type: type,
      stats: stats,
      text: text,
      element: _elementFromString(element),
      cost: costMap,
      image: image,
    ));
  }

  /// Clear all cards from the generator.
  void clearCards() {
    cards = <GeneratorCard>[];
  }

  /// Export the current document.
  Future<List<int>> saveDocument() async {
    var _document = Document(title: title);

    var chunks = [];

    for (var i = 0; i < cards.length; i += 9) {
      var end = (i + 9 < cards.length) ? i + 9 : cards.length;
      chunks.add(cards.sublist(i, end));
    }

    await chunks.forEach((var pageCards) async =>
        _document.addPage(await _buildPage(_document, pageCards)));

    return _document.save();
  }

  /// Export the current document as single cards.
  Future<List<int>> saveSingleCards() async {
    var _document = Document(title: title);

    await cards.forEach((var card) async =>
        _document.addPage(await _buildSingleCardPage(_document, card)));

    return _document.save();
  }

  /// Builds a single page of up to 9 cards.
  Future<Page> _buildPage(Document document, List<GeneratorCard> cards) async {
    var cardWidgets = List.from(
        await Future.wait(cards
            .map((GeneratorCard card) => _buildCardWidget(document, card))),
        growable: true);

    while (cardWidgets.length < 9) {
      cardWidgets.add(
        SizedBox(
          height: cardHeight,
          width: cardWidth,
        ),
      );
    }
    return Page(
      pageFormat: a4Format,
      build: (Context context) {
        return GridView(
          crossAxisCount: 3,
          children: cardWidgets,
        );
      },
    );
  }

  /// Builds a page from a single card.
  Future<Page> _buildSingleCardPage(
      Document document, GeneratorCard card) async {
    var cardWidget = await _buildCardWidget(document, card);

    return Page(
      pageFormat: a4Format,
      build: (Context context) {
        return cardWidget;
      },
    );
  }

  /// Builds a Widget for a single card.
  Future<Widget> _buildCardWidget(Document document, GeneratorCard card) async {
    return SizedBox(
      height: cardHeight,
      width: cardWidth,
      child: Stack(
        children: [
          Image(
            await _buildImage(document, frameImages[card.element]),
          ),
          Stack(
            children: [
              Positioned(
                  left: 1.9 * mm,
                  top: 2.6 * mm,
                  child: Text(card.name, style: styleTitle)),
              Positioned(
                  left: 56 * mm,
                  top: 1.5 * mm,
                  child: SizedBox(
                      height: 6 * mm,
                      width: 6 * mm,
                      child: Image(await _buildImage(
                          document, elementImages[card.element])))),
              Positioned(
                  left: 1 * mm,
                  top: 8 * mm,
                  child: SizedBox(
                    height: 45 * mm,
                    width: 61.5 * mm,
                    child: Center(
                      child: Image(
                        await _buildImage(document, card.image),
                      ),
                    ),
                  )),
              Positioned(
                  left: 2 * mm,
                  top: 54.1 * mm,
                  child: Text(card.type, style: styleType)),
              Positioned(
                right: 2 * mm,
                top: 54.1 * mm,
                child: Text(card.id, style: styleId),
              ),
              Positioned(
                  left: 4.1 * mm,
                  top: 58.5 * mm,
                  child: SizedBox(
                      height: 24.1 * mm,
                      width: 55.4 * mm,
                      child: await _buildCardText(document, card.text))),
              Positioned(
                  left: 1.5 * mm,
                  bottom: 1.5 * mm,
                  child: Row(
                      mainAxisAlignment: MainAxisAlignment.start,
                      crossAxisAlignment: CrossAxisAlignment.end,
                      children: await _buildCostRow(document, card.cost))),
              Positioned(
                  right: 2 * mm,
                  bottom: 1.2 * mm,
                  child: Row(
                      mainAxisAlignment: MainAxisAlignment.end,
                      crossAxisAlignment: CrossAxisAlignment.end,
                      children: [Text(card.stats, style: styleStats)]))
            ],
          ),
        ],
      ),
    );
  }

  /// Builds the cost row Widgets for a card.
  Future<List<Widget>> _buildCostRow(
      Document document, Map<GeneratorElement, int> cost) async {
    var a = cost.keys.map((key) async {
      return await Future.wait(
        List.generate(cost[key], (int i) async {
          return SizedBox(
            height: 4 * mm,
            width: 4 * mm,
            child: Image(
              await _buildImage(document, elementImages[key]),
            ),
          );
        }),
      );
    });
    var b = await Future.wait(a);
    return b.expand((pair) => pair).toList();
  }

  /// Builds card text.
  Future<RichText> _buildCardText(Document document, String text) async {
    var result = <InlineSpan>[];

    var p = 0;

    for (var i = 0; i < text.length; i++) {
      if (text[i] == '<' || text[i] == '[') {
        if (i > p) {
          result.add(TextSpan(text: text.substring(p, i)));
        }
        p = i + 1;
      }
      if (text[i] == '>') {
        if (i > p) {
          result.add(TextSpan(text: text.substring(p, i), style: styleKeyword));
        }
        p = i + 1;
      }
      if (text[i] == ']') {
        if (i > p) {
          result.add(
            WidgetSpan(
              baseline: -0.3 * mm,
              child: Image(
                await _buildImage(
                  document,
                  elementImages[
                      _elementFromString(text.substring(p, i).toLowerCase())],
                ),
              ),
            ),
          );
        }
        p = i + 1;
      }
    }
    if (text.length > p) {
      result.add(TextSpan(text: text.substring(p, text.length)));
    }

    return RichText(text: TextSpan(children: result, style: styleText));
  }

  /// Builds a single image for a document.
  Future<PdfImage> _buildImage(Document document, img.Image image) async {
    return PdfImage(document.document,
        image: image.data.buffer.asUint8List(),
        width: image.width,
        height: image.height);
  }

  /// Converts a string to a GeneratorElement.
  GeneratorElement _elementFromString(String string) {
    switch (string) {
      case 'fire':
        return GeneratorElement.fire;
        break;
      case 'earth':
        return GeneratorElement.earth;
        break;
      case 'air':
        return GeneratorElement.air;
        break;
      case 'water':
        return GeneratorElement.water;
        break;
      case 'spirit':
        return GeneratorElement.spirit;
        break;
      case 'any':
      case 'neutral':
        return GeneratorElement.neutral;
        break;
      default:
        throw FallThroughError();
    }
  }
}
